/*
 * NAND defines for DaVinci Initial Bootloader.
 *
 * Copyright (c) 2007 Sergey Kubushyn <ksi@koi8.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */
#ifndef _NAND_H_
#define _NAND_H_

#include <config.h>
#include <sys/types.h>

#define	TINY_PAGE	256
#define SMALL_PAGE	512
#define LARGE_PAGE	2048

#define NAND_DATA_ADDR  ((volatile void *)(NAND_BASE))
#define NAND_ALE_ADDR	((volatile void *)(NAND_BASE + 0x08))
#define NAND_CMD_ADDR	((volatile void *)(NAND_BASE + 0x10))

#define	NAND_READ_ID		0x90
#define NAND_READ_ID_ADDR	0x00

#define DV_NANDFCR		((u_int32_t *)(0x01e00060))
#define DV_NANDFSR		((u_int32_t *)(0x01e00064))
#define DV_NANDF1ECC		((u_int32_t *)(0x01e00070))
#define DV_AB1CR		((u_int32_t *)(0x01e00010))

#define NAND_TIMEOUT		0x3000

#define	NAND_STATUS		0x70
#define NAND_STATUS_BSY		0x40
#define NAND_STATUS_WP		0x80
#define NAND_STATUS_ERR		0x01

#define NAND_LO_PAGE		0
#define NAND_READ_0x30		0x30
#define	NAND_PGM_START		0x80
#define NAND_PGM_END		0x10
#define	NAND_BERASE1		0x60
#define	NAND_BERASE2		0xd0
#define	NAND_RST		0xff

#define UNKNOWN_NAND		0xff

#define NAND_LDR_MAGIC		0xa1aced00
#define NAND_UA_MAGIC		0x64686175

#define NAND_LDR_START_BLK	1
#define NAND_LDR_TRY_BLKS	5

#define NAND_UA_START_BLK	6
#define NAND_UA_TRY_BLKS	5

#define NAND_UA_LOAD_FAIL	0xffffffff

typedef struct {
	u_int8_t	dev_id;		/* DevID from device */
	u_int8_t	page_cnt;	/* Pages per block */
	u_int16_t	page_size;	/* Bytes per page, including OOB */
	u_int8_t	blk_shift;	/* How many bytes to shift block address */
	u_int8_t	addr_cycles;	/* Number of address cycles */
} nand_dev_info;

typedef struct {
	u_int8_t	nand_width;	/* 1 - 16 bit, 0 - 8 bit */
	u_int8_t	dev_id;		/* NAND device ID */
	u_int8_t	page_cnt;	/* Pages per block */
	u_int8_t	addr_cycles;	/* No. of address cycles */
	u_int8_t	is_large_page;	/* 1 - large page, 0 - small/tiny */
	u_int16_t	page_size;	/* Bytes per page, including OOB */
	u_int8_t	blk_shift;	/* Block address shift */
	u_int8_t	page_shift;	/* Page address shift */
} nand_dev;

typedef struct {
	u_int32_t	magic;		/* Magic number */
	u_int32_t	entry;		/* Loader entry point address */
	u_int32_t	num_pages;	/* Number of pages loader occupies */
	u_int32_t	block;		/* Starting block of loader code */
	u_int32_t	page;		/* Starting page of loader code */
} nand_ldr_header;

typedef struct {
	u_int32_t	magic;		/* Magic number */
	u_int32_t	entry;		/* UserApp (UA) entry point address */
	u_int32_t	num_pages;	/* Number of pages UA occupies */
	u_int32_t	base_load_addr;	/* UA bulk load address */
	u_int32_t	blk_no;		/* Current block sequence number */
	u_int32_t	blk_start_page; /* Starting page with UA in this block */
	u_int32_t	blk_last_page;	/* Last page with UA in this block */
	u_int32_t	blk_load_addr;	/* This block data load address */
} nand_ua_hdr;

/* Assuming NAND on CS2 */
static inline void nand_ecc_start(void) {
	*(volatile u_int32_t *)(DV_NANDFCR) |= (1 << 8);
}

static inline u_int32_t nand_ecc_read(void) {
	 return *(volatile u_int32_t *)(DV_NANDF1ECC);
}

static inline void nand_write_cmd(u_int8_t cmd)
{
	*(volatile u_int8_t *)NAND_CMD_ADDR = cmd;
}

static inline void nand_write_addr(u_int8_t addr)
{
	*(volatile u_int8_t *)NAND_ALE_ADDR = addr;
}

static inline void nand_write_data8(u_int8_t data)
{
	*(volatile u_int8_t *)NAND_DATA_ADDR = data;
}

static inline void nand_write_data16(u_int16_t data)
{
	*(volatile u_int16_t *)NAND_DATA_ADDR = data;
}

static inline u_int8_t nand_read_data8(void)
{
	return(*((volatile u_int8_t *)NAND_DATA_ADDR));
}

static inline u_int16_t nand_read_data16(void)
{
	return(*((volatile u_int16_t *)NAND_DATA_ADDR));
}

int nand_init(void);
int nand_ldr_write(void);
int nand_ua_write(u_int8_t *buf, u_int32_t size, u_int32_t entry, u_int32_t load_addr);
int nand_ua_load(void);


#define NAND_SEND_ADDR	\
	switch (nand_info.addr_cycles) {\
		case 5:\
			addr = (blk_addr << (nand_info.blk_shift - 16)) | page;\
			nand_write_addr(0x00);\
			nand_write_addr(0x00);\
			nand_write_addr(addr & 0xff);\
			nand_write_addr((addr >> 8) & 0xff);\
			nand_write_addr((addr >> 16) & 0xff);\
			break;\
		case 4:\
			addr = (blk_addr << nand_info.blk_shift) | (page << nand_info.page_shift);\
			nand_write_addr(addr & 0xff);\
			nand_write_addr((addr >> 8) & 0xff);\
			nand_write_addr((addr >> 16) & 0xff);\
			nand_write_addr((addr >> 24) & 0xff);\
			break;\
		default:\
			addr = (blk_addr << nand_info.blk_shift) | (page << nand_info.page_shift);\
			nand_write_addr(addr & 0xff);\
			nand_write_addr((addr >> 8) & 0xff);\
			nand_write_addr((addr >> 16) & 0xff);\
	}
#endif
