/*
 * Main program for DaVinci Initial Bootloader.
 *
 * Copyright (c) 2007 Sergey Kubushyn <ksi@koi8.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <config.h>
#include <loader.h>
#include <nand.h>

#define NAND_INIT	\
		printstr(nand_init_str);\
		if (nand_init()) {\
			printstr(fail_str);\
			goto beg;\
		}\
		printstr(ok_str);


void menu(void)
{
	char *menu_str = \
	"\r\n\r\n------------------------------------------------------\r\n" \
	"           KSI Boot Loader v." VERSION "\r\n" \
	"------------------------------------------------------\r\n" \
	"L addr size           - Load raw binary file to memory\r\n" \
	"G addr                - Branch to addr\r\n" \
	"A buf size entry load - Write user app to NAND\r\n" \
	"N buf size            - Write user app to NOR\r\n" \
	"U                     - Write Boot Loader to NAND\r\n" \
	"D start count         - Dump memory\r\n" \
	"F start count value   - Fill memory (bytes)\r\n" \
	"I start count value   - Fill memory (ints)\r\n" \
	"------------------------------------------------------\r\n\r\n" \
	"ldr> ";

	printstr(menu_str);
}

void uboot2nor(void)
{
	printstr("\r\nUser App -> NOR... Not implemented yet.\r\n");
}

int ascii2int(u_int8_t *asc)
{
	int	i, tmp, result;
	
	result = 0;
	
	for (i = 0; i < 8; i++) {
		tmp = 0x000000ff & (asc[i] > '9' ? asc[i] & 0xdf : asc[i]);
		result |= ((tmp <= '9') ? (tmp - 0x30) : (tmp - 0x37)) << (4 * (7 - i));
	}

	return(result);
}

void int2ascii(int data, char *asc)
{
	int	i;
	char	tmp;
	
	for (i = 0; i < 8; i++) {
		tmp = (data >> (4 * (7 - i))) & 0x0000000f;
		asc[i] = tmp > 0x09 ? tmp + 0x37 : tmp + 0x30;
	}
}

void byte2ascii(u_int8_t data, char *asc)
{
	char	tmp;
	
	tmp = (data >> 4) & 0x0f;
	asc[0] = tmp > 0x09 ? tmp + 0x37 : tmp + 0x30;
	tmp = data & 0x0f;
	asc[1] = tmp > 0x09 ? tmp + 0x37 : tmp + 0x30;
}

int is_valid_hex(u_int8_t dig)
{
	if ((dig < '0') || (dig > 'f'))
		return(0);
	if ((dig > '9') && (dig < 'A'))
		return(0);
	if ((dig > 'F') && (dig < 'a'))
		return(0);
	return(1);
}

#define CTRL_C_STR	0x80
#define EMPTY_STR	0x00
#define BAD_STR		0xff

int get_cmd(u_int32_t **args)
{
	u_int8_t	tmp[8];
	int		i, len, cur_pos, args_no, in_arg;

	if ((len = getstr()) < 0)	/* ^C, start over */
		return(CTRL_C_STR);
	else if (len == 0)		/* Empty string */
		return(EMPTY_STR);

	args_no = 0;
	for (i = 0; i < 8; i++)
		tmp[i] = '0';
		
	/* First, find the command itself (must be one letter only) */
	for (cur_pos = 0; cur_pos < len; cur_pos++) {
		if (__str_buf[cur_pos] != ' ')
			break;
	}
	
	if (cur_pos >= len)
		return(EMPTY_STR);
	else
		*(args[0]) = __str_buf[cur_pos++];
	
		
	if ((cur_pos == len) || (__str_buf[cur_pos] == ' '))
		args_no++;
	else
		return(BAD_STR);

	in_arg = i = 0;
	while (cur_pos <= len) {
		if ((__str_buf[cur_pos] == ' ') || (cur_pos == len)) {
			if (!in_arg) {
				cur_pos++;
				continue;
			} else {
				if ((i > 8) || (args_no > 4))
					return(BAD_STR);
					
				/* Get argument */
				for (; i > 0; i--)
					tmp[8 - i] = __str_buf[cur_pos - i];
				
				*(args[args_no++]) = ascii2int(tmp);

				for (; i < 8; i++)
					tmp[i] = '0';
					
				i = 0;
				in_arg = 0;
				cur_pos++;
			}
		} else {
			if (!is_valid_hex(__str_buf[cur_pos]))
				return(BAD_STR);

				in_arg = 1;
			i++;
			cur_pos++;
		}
	}
	
	return(args_no);
}

void dump(u_int32_t start, u_int32_t count)
{
	int		i;
	char		buf[11];
	u_int8_t	*ptr = (u_int8_t *)(start & 0xfffffff0);

	printstr("\r\n");
	buf[8] = buf[9] = ' ';
	buf[10] = 0;

	for (i = 0; i < count; i++) {
		if ((i & 0x0000000f) == 0) {
			printstr("\r\n");
			int2ascii((u_int32_t)ptr, buf);
			printstr(buf);
		}
		else if ((i & 0x00000003) == 0) {
			printstr(" ");
		}

		buf[0] = ' ';
		byte2ascii(*ptr++, &buf[1]);
		buf[3] = 0;
		printstr(buf);
	}

	printstr("\r\n");
}

void fill_byte(u_int32_t start, u_int32_t count, u_int8_t value)
{
	int		i;
	u_int8_t	*ptr = (u_int8_t *)start;

	for (i = 0; i < count; i++) {
		*ptr++ = value;
	}
}

void fill_int(u_int32_t start, u_int32_t count, u_int32_t value)
{
	int		i;
	u_int32_t	*ptr = (u_int32_t *)start;

	for (i = 0; i < count; i++) {
		*ptr++ = value;
	}
}

int main(void)
{
	int		nand_init_done;
	int		arg_no;
	u_int32_t	arg0, arg1, arg2, arg3, arg4;
	u_int32_t *	argv[5] = {&arg0, &arg1, &arg2, &arg3, &arg4};
	char		*what_str = "\r\nSorry, WHAT?\r\n";
	char		*nand_init_str = "\r\nNAND init ";
	char		*fail_str = "FAILED!\r\n";
	char		*ok_str = "OK.\r\n";
	char		*go_str = "\r\nGO_AHEAD\r\n";
	char		*got_it_str = "\r\nGOT_IT\r\n";
	char		*writing_str = "\r\nWriting ";
	char		*loader_str = "bootloader to ";
	char		*app_str = "user app to ";
	char		*nand_str = "NAND... ";
	char		*nor_str = "NOR... ";

	if (!((*BOOTCFG) & 0x000000c0)) {
		/* NAND Boot */
		if (!nand_init())
			if ((arg0 = nand_ua_load()) != NAND_UA_LOAD_FAIL)
				jump(arg0);
	}

beg:
	menu();

	/* Get user input and parse it */
	if ((arg_no = get_cmd(argv)) == CTRL_C_STR)
		goto beg;
	else if ((arg_no == EMPTY_STR) || (arg_no == BAD_STR))
		goto what;
	
	switch (arg0) {
		case 'D':
		case 'd':
			if (arg_no != 3)
				goto what;
			dump(arg1, arg2);
			goto beg;
		case 'F':
		case 'f':
			if (arg_no != 4)
				goto what;
			fill_byte(arg1, arg2, (u_int8_t)(arg3 & 0x000000ff));
			goto beg;
		case 'I':
		case 'i':
			if (arg_no != 4)
				goto what;
			fill_int(arg1, arg2, arg3);
			goto beg;
		case 'L':
		case 'l':
			if (arg_no != 3)
				goto what;
			printstr(go_str);
			load_raw(arg1, arg2);
			printstr(got_it_str);
			goto beg;
		case 'G':
		case 'g':
			if (arg_no != 2)
				goto what;
			jump(arg1);
		case 'A':
		case 'a':
			if (arg_no != 5)
				goto what;
			NAND_INIT
			printstr(writing_str);
			printstr(app_str);
			printstr(nand_str);
			if (nand_ua_write((u_int8_t *)arg1, arg2, arg3, arg4))
				printstr(fail_str);
			else
				printstr(ok_str);
			goto beg;
		case 'N':
		case 'n':
#if 0	/* Not implemented yet */
			if (arg_no != 3)
				goto what;
			NOR_INIT
			printstr(writing_str);
			printstr(app_str);
			printstr(nor_str);
			if (nor_ua_load() == NOR_UA_LOAD_FAIL)
				printstr(fail_str);
			else
				printstr(ok_str);
			goto beg;
#else
			uboot2nor();
			goto beg;
#endif
		case 'U':
		case 'u':
			NAND_INIT
			printstr(writing_str);
			printstr(loader_str);
			printstr(nand_str);
			if (nand_ldr_write())
				printstr(fail_str);
			else
				printstr(ok_str);
			goto beg;
		default:
			break;
	}

what:
	printstr(what_str);
	goto beg;

	return 0;
}
