/*
 * NAND Routines for DaVinci Initial Bootloader.
 *
 * Copyright (c) 2007 Sergey Kubushyn <ksi@koi8.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <config.h>
#include <loader.h>
#include <nand.h>

nand_dev_info nand_info_tbl[] = {
	{0x73, 32, SMALL_PAGE, 13, 3},	/* 16 MiB */
	{0x33, 32, SMALL_PAGE, 13, 3},	/* 16 MiB */
	{0x75, 32, SMALL_PAGE, 13, 3},	/* 32 MiB */
	{0x35, 32, SMALL_PAGE, 13, 3},	/* 32 MiB */
	{0x43, 32, SMALL_PAGE, 12, 3},	/* 16 MiB */
	{0x45, 32, SMALL_PAGE, 12, 3},	/* 32 MiB */
	{0x53, 32, SMALL_PAGE, 12, 3},	/* 16 MiB */
	{0x55, 32, SMALL_PAGE, 12, 3},	/* 32 MiB */
	{0x36, 32, SMALL_PAGE, 13, 4},	/* 64 MiB */
	{0x79, 32, SMALL_PAGE, 13, 4},	/* 128 MiB */
	{0x71, 32, SMALL_PAGE, 13, 4},	/* 256 MiB */
	{0x46, 32, SMALL_PAGE, 13, 4},	/* 64 MiB */
	{0x56, 32, SMALL_PAGE, 13, 4},	/* 64 MiB */
	{0x74, 32, SMALL_PAGE, 13, 4},	/* 128 MiB */
	{0xa1, 64, LARGE_PAGE, 22, 4},	/* 128 MiB */
	{0xaa, 64, LARGE_PAGE, 22, 5},	/* 256 MiB */
	{0xda, 64, LARGE_PAGE, 22, 5},	/* 256 MiB */
	{0xdc, 64, LARGE_PAGE, 22, 5},	/* 512 MiB */
	{0x76, 32, SMALL_PAGE, 13, 4},	/* 64 MiB */
	{0xac, 64, LARGE_PAGE, 22, 5},	/* 512 MiB */
	{0xb1, 64, LARGE_PAGE, 22, 4},	/* 128 MiB */
	{0xc1, 64, LARGE_PAGE, 22, 4},	/* 128 MiB */
	{0xf1, 64, LARGE_PAGE, 22, 4},	/* 128 MiB */
	{0x00}				/* End of table */
};


nand_dev		nand_info;
nand_ldr_header		nand_boot;


void printint(int data)
{
	char	tmp[8];
	int	i;
	
	int2ascii(data, tmp);
	for (i = 0; i < 8; i++)
		putchr(tmp[i]);
}


/* Assuming NAND on CS2 */
void nand_emif_cfg(void)
{
    if (nand_info.nand_width) {
	  *DV_AB1CR = 0x3ffffffd;
    } 

    *DV_NANDFCR = 0x1; 
}


void nand_read_bytes(u_int32_t num_bytes, u_int8_t *dst_addr)
{
	int	i;

	i = nand_ecc_read();
	nand_ecc_start();

	if (nand_info.nand_width) {
		/* 16-bit NAND */
		for(i = 0; i < (num_bytes / 2); i++) {
			*(u_int16_t *)dst_addr = nand_read_data16();
			dst_addr += 2;
		}
	} else {
		/* 8-bit NAND */
		for (i = 0; i < num_bytes; i++) {
			*dst_addr = nand_read_data8();
			dst_addr++;
		}
	}
}


/* Get NAND info */
int nand_get_info(void)
{
	int	i, dev_id;

	nand_write_cmd(NAND_READ_ID);
	nand_write_addr(NAND_READ_ID_ADDR);

	wait_loop(10);

	/* Throw away Manufacturer ID and extra ID, just read Device ID */
	i = nand_read_data8();
	dev_id = nand_read_data8();
	i = nand_read_data8();
	i = nand_read_data8();

	/* Search for the NAND in nand_info_tbl */
	i = nand_info.dev_id = 0;
	while (nand_info_tbl[i].dev_id) {
		if (dev_id != nand_info_tbl[i].dev_id) {
			i++;
			continue;
		}

		nand_info.dev_id = dev_id;
		nand_info.page_cnt = nand_info_tbl[i].page_cnt;
		nand_info.page_size = nand_info_tbl[i].page_size;
		nand_info.blk_shift = nand_info_tbl[i].blk_shift;
		nand_info.addr_cycles = nand_info_tbl[i].addr_cycles;

		if (nand_info.page_size > SMALL_PAGE) {
			nand_info.page_shift = 16;
			nand_info.is_large_page = 1;
		} else {
			nand_info.page_shift = 8;
			nand_info.is_large_page = 0;
		}
		
		break;
	}

	if (!nand_info.dev_id) {
		nand_info.dev_id = UNKNOWN_NAND;
		nand_info.page_cnt = 0;
		nand_info.page_size = 0;
		nand_info.blk_shift = 0;
		nand_info.addr_cycles = 0;
		nand_info.page_shift = 0;
		nand_info.is_large_page = 0;
		return(1);
	} else {
		return(0);
	}
}


int nand_wait_rdy(void)
{
	int	cnt;

	for (cnt = 0; cnt < NAND_TIMEOUT; cnt++) {
		if (*DV_NANDFSR & 0x1)
			return(0);
			
		wait_loop(100);
	}

	return(1);
}


int nand_check_status(void)
{
	nand_write_cmd(NAND_STATUS);
	return(nand_read_data8() & NAND_STATUS_ERR);
}


u_int32_t nand_oob_ecc_read(u_int8_t *read_ptr)
{
	u_int32_t oob_ecc_val;

	if(nand_info.nand_width)
		/* 16 bit NAND */
		oob_ecc_val = (*((u_int16_t *)(&read_ptr[0])) << 16) |
			(*((u_int16_t *)(&read_ptr[2]))); 
	else
		/* 8 bit NAND */
		oob_ecc_val = (read_ptr[0] << 24) | (read_ptr[1] << 16) |
			(read_ptr[2] << 8) | (read_ptr[3]); 
    	
	return(oob_ecc_val);
}


int nand_read_page(u_int32_t blk_addr, u_int32_t page, u_int8_t *dest)
{
	u_int32_t	i;
	u_int32_t	addr;
	u_int32_t	ecc_value[4];
	u_int32_t	oob_value;


	nand_write_cmd(NAND_LO_PAGE);

	NAND_SEND_ADDR

	if (nand_info.is_large_page)	/* Large Page NAND */
		nand_write_cmd(NAND_READ_0x30);

	wait_loop(10);

	if (nand_wait_rdy())
		return(1);

	/* Read data to the destination buffer and detect errors */
	switch (nand_info.page_size) {
		case SMALL_PAGE:
			nand_read_bytes(nand_info.page_size, dest);
			ecc_value[0] = nand_ecc_read() & 0x0fff0fff;
			nand_read_bytes(16, &dest[nand_info.page_size]);
			oob_value = nand_oob_ecc_read(&dest[nand_info.page_size]) & 0x0fff0fff;
			if (ecc_value[0] != oob_value) {
				return(1);
			}
			break;
		case LARGE_PAGE:
			for ( i = 0; i < 4; i++) {
				nand_read_bytes(512, &dest[512 * i]);
				ecc_value[i] = nand_ecc_read() & 0x0fff0fff;
			}
			nand_read_bytes(64, &dest[nand_info.page_size]);

			for(i = 0; i < 4; i++) {
				/* ECC values read from (2056 + i * 16 + 8) */
				oob_value = nand_oob_ecc_read(&dest[nand_info.page_size + 8 + (i * 16)]) & 0xfff0fff;
				if(ecc_value[i] != oob_value)
					return(1);
			}
			
			break;
		default:
			return(1);
	}

	wait_loop(100);

	if (nand_wait_rdy())
		return(1);

	return(0);
}


int nand_write_page(u_int32_t blk_addr, u_int32_t page, u_int8_t *src)
{
	int		i,j;
	u_int32_t	addr;
	u_int32_t	ecc_value[4];

	i = nand_ecc_read();
	nand_ecc_start();

	nand_write_cmd(NAND_PGM_START);

	NAND_SEND_ADDR

	i = nand_ecc_read();
	nand_ecc_start();

	switch (nand_info.page_size) {
		case SMALL_PAGE:
			if (nand_info.nand_width) {
				/* 16-bit NAND */
				for (i = 0; i < nand_info.page_size; i += 2)
					nand_write_data16(*((u_int16_t *)&src[i]));
				ecc_value[0] = (nand_ecc_read() & 0x0fff0fff);
				nand_write_data16((ecc_value[0] >> 16) & 0xffff);
				nand_write_data16(ecc_value[0] & 0xffff);
				for (i = 0; i < 6; i++)
					nand_write_data16(0xffff);
			} else {
				/* 8-bit NAND */
				for (i = 0; i < nand_info.page_size; i++)
					nand_write_data8(src[i]);
				ecc_value[0] = (nand_ecc_read() & 0x0fff0fff);
				nand_write_data8((ecc_value[0] >> 24) & 0xff);
				nand_write_data8((ecc_value[0] >> 16) & 0xff);
				nand_write_data8((ecc_value[0] >> 8) & 0xff);
				nand_write_data8(ecc_value[0] & 0xff);
				for (i = 0; i < 12; i++)
					nand_write_data8(0xff);
			}
			break;
		case LARGE_PAGE:
			j = 0;
			if (nand_info.nand_width) {
				/* 16-bit NAND */
				for (i = 0; i < nand_info.page_size; i += 2) {
					nand_write_data16(*((u_int16_t *)&src[i]));
					if ((i == 510) || (i == 1022) || (i == 1534) || (i == 2046)) {
						ecc_value[j] = nand_ecc_read() & 0x0fff0fff;
						nand_ecc_start();
						j++;
					}
				}
				for (j = 0; j < 4; j++) {
					for (i = 0; i < 4; i++)
						nand_write_data16(0xffff);
					nand_write_data16((ecc_value[j] >> 16) & 0xffff);
					nand_write_data16(ecc_value[j] & 0xffff);
					for (i = 0; i < 2; i++)
						nand_write_data16(0xffff);
				}
			} else {
				/* 8-bit NAND */
				for (i = 0; i < nand_info.page_size; i++) {
					nand_write_data8(src[i]);

					if ((i == 511) || (i == 1023) || (i == 1535) || (i == 2047)) {
						ecc_value[j] = nand_ecc_read() & 0x0fff0fff;
						nand_ecc_start();
						j++;
					}
				}
				for (j = 0; j < 4; j++) {
					for (i = 0; i < 8; i++)
						nand_write_data8(0xff);
					nand_write_data8((ecc_value[j] >> 24) & 0xff);
					nand_write_data8((ecc_value[j] >> 16) & 0xff);
					nand_write_data8((ecc_value[j] >> 8) & 0xff);
					nand_write_data8(ecc_value[j] & 0xff);
					for (i = 0; i < 4; i++)
						nand_write_data8(0xff);
				}
			}
			break;
		default:
			return(1);
	}

	nand_write_cmd(NAND_PGM_END);

	wait_loop(10000);

	if (!nand_wait_rdy())
		return(nand_check_status());

	return(1);
}

/* NAND Erase function */
int nand_erase_block(u_int32_t blk_addr)
{
	u_int32_t	addr = blk_addr << (nand_info.blk_shift - nand_info.page_shift);

	nand_write_cmd(NAND_BERASE1);
	nand_write_addr(addr & 0xff);
	nand_write_addr((addr >> 8) & 0xff);

	if(nand_info.addr_cycles == 4) {
		nand_write_addr((addr >> 16) & 0xff);
	}

	nand_write_cmd(NAND_BERASE2);

	wait_loop(10000);

	if (!nand_wait_rdy())
		return(nand_check_status());

	return(1);
}

int nand_init(void) {
	int	rc;

	/* Set nand_width (0 - 8 bit NAND, 1 - 16 bit). AEMIF CS2 bus width
	 * is given by BOOTCFG(bit 5)
	 */
	nand_info.nand_width = (*BOOTCFG & 0x00000020) >> 5;

	/* initialize NAND interface */
	nand_emif_cfg();

	/* Send reset command to NAND */
	nand_write_cmd(NAND_RST);

	/* Wait for reset to complete */
	if (nand_wait_rdy()) {
		rc = 1;
	} else {
		/* Identify the NAND used */
		rc = nand_get_info();
	}

	return(rc);
}


/* Find User Application (UA) and load it to DRAM (or whatever) */
int nand_ua_load(void) {
	u_int8_t	*app_code;
	u_int8_t	*hdr_buf;
	u_int8_t	*page_buf;
	nand_ua_hdr	*hdr;
	u_int32_t	block, blk_seq, page;
	int		i, rc, num_pages, blk_attempts;

	hdr_buf = (u_int8_t *)(NAND_BUF_START);
	hdr = (nand_ua_hdr *)&hdr_buf[0];
	page_buf = (u_int8_t *)(NAND_BUF_START - NAND_BUF_SIZE);

	printstr("\r\nSearching for UserApp in NAND... ");

	block = NAND_UA_START_BLK;
	blk_seq = 0;

	/* Find and read initial UA header */
	for (i = 0; i < NAND_UA_TRY_BLKS; i++, block++) {
		if (nand_read_page(block, 0, hdr_buf)) {
			hdr->magic = 0xdeaddead;
			continue;
		}

		if (hdr->magic == NAND_UA_MAGIC)
			break;
	}
	
	if ((hdr->magic != NAND_UA_MAGIC) || (hdr->blk_no != blk_seq))
		goto nuald_fail;

	printstr("Found.\r\nChecking header... ");
	
	/* Yes, this should be here -- we're not linking libgcc in */
	switch (nand_info.page_size) {
		case SMALL_PAGE:
			page = DRAM_SIZE >> 9;
			break;
		default:
			page = DRAM_SIZE >> 11;
	}

	if (	(hdr->num_pages == 0) ||
		(hdr->num_pages > page) ||
		(hdr->base_load_addr < DRAM_START) ||
		(hdr->base_load_addr != hdr->blk_load_addr) ||
		((hdr->base_load_addr + nand_info.page_size * hdr->num_pages) > NAND_BUF_START)
	   )
	   	goto nuald_fail;
	   	
	printstr("OK.\r\nLoading UserApp from NAND.");

	app_code = (u_int8_t *)(hdr->base_load_addr);
	num_pages = hdr->num_pages;
	blk_attempts = NAND_UA_TRY_BLKS;

	/* OK, here ist da Big Main Loop :) */
	while (1) {
		/* Read current block pages with UA code */
		for (page = hdr->blk_start_page; page <= hdr->blk_last_page; page++) {
			if (rc = nand_read_page(block, page, page_buf))
				break;
			
			for (i = 0; i < nand_info.page_size; i++)
				*app_code++ = *(page_buf + i);
/* KSI: Beautification -- put '+' at the start of a block !!! */
			putchr('.');
		}
		
		if (!rc) {
			/* OK, block is read, go find the next one
			 * if not all pages are read in */
			num_pages -= hdr->blk_last_page - hdr->blk_start_page + 1;
			if (num_pages <= 0)
				break;
			blk_seq++;
			blk_attempts = NAND_UA_TRY_BLKS;
		} else {
			/* Error reading block, go find a spare one */
			if(--blk_attempts <= 0) {
				/* Attempts limit exhausted, fail */
				goto nuald_fail;
			}
		}

		block++;

nuald_nxt_blk:
		for (i = blk_attempts; i > 0; block++, i--) {
			if (nand_read_page(block, 0, hdr_buf)) {
				hdr->magic = 0xdeaddead;
				continue;
			}
			
			if (hdr->magic == NAND_UA_MAGIC)
				break;
		}
		
		/* OK, we've finished searching, check if a valid block */
		if (hdr->magic != NAND_UA_MAGIC)
			goto nuald_fail;

		/* Now check if the sequence is right, skip if already read in */
		if (hdr->blk_no == blk_seq) {
			blk_attempts = i;
			app_code = (u_int8_t *)(hdr->blk_load_addr);
			continue;
		} else if (hdr->blk_no > blk_seq) {
			goto nuald_fail;
		} else {
			if (!rc) {
				if ((blk_seq - hdr->blk_no) == 1) {
					block++;
					goto nuald_nxt_blk;
				}
			}
nuald_fail:
				printstr("FAILED!\r\n");
				return(NAND_UA_LOAD_FAIL);
		}

	}

	printstr("Done.\r\n");

	return(hdr->entry);
}

void nand_wp_set(int onoff)
{
#ifdef SONATA_BOARD_GPIOWP
	if (onoff)
		*DV_GPIO_CLR_DATA01 = GPIO_MASK(NAND_WP_GPIO);
	else
		*DV_GPIO_SET_DATA01 = GPIO_MASK(NAND_WP_GPIO);
#endif
}


/* KSI: Beautification -- put dots and + for a page written !!! */
int nand_ua_write(u_int8_t *buf, u_int32_t size, u_int32_t entry, u_int32_t load_addr)
{
	u_int8_t	*app_code, *ptr;
	u_int8_t	*hdr_buf, *page_buf;
	nand_ua_hdr	*hdr;
	int		i, block, page, page_count;
	
	hdr_buf = (u_int8_t *)(NAND_BUF_START);
	hdr = (nand_ua_hdr *)&hdr_buf[0];
	page_buf = (u_int8_t *)(NAND_BUF_START - NAND_BUF_SIZE);

	/* First of all, build the initial header */
	for (i = 0; i < NAND_BUF_SIZE; i++)
		hdr_buf[i] = 0xff;
		
	/* 
	 * Now a little bit of arithmetic. No, we can't just
	 * do something like "size / nand_info.page_size".
	 */
	switch (nand_info.page_size) {
		case SMALL_PAGE:
			page_count = size / SMALL_PAGE;
			if (size & 0x1ff)
				page_count++;
			break;
		case LARGE_PAGE:
			page_count = size / LARGE_PAGE;
			if (size & 0x7ff)
				page_count++;
			break;
		default:
			return(1);	/* Won't work with tiny page NAND */
	}

	hdr->num_pages = page_count;
	
	hdr->blk_last_page = page_count < (nand_info.page_cnt - 1) ?
		page_count : nand_info.page_cnt - 1;

	hdr->magic = NAND_UA_MAGIC;
	hdr->entry = entry;
	hdr->base_load_addr = load_addr;
	hdr->blk_no = 0;
	hdr->blk_start_page = 1;
	hdr->blk_load_addr = load_addr;

	/* OK, initial header is ready, let's start the show... */
	nand_wp_set(0);
	block = NAND_UA_START_BLK;
	app_code = buf;
	i = 0;
	
	/* Main Loop */
	while (1) {
		for (; i < NAND_UA_TRY_BLKS; i++) {
			if (!nand_erase_block(block))
				break;
		
			block++;
		}

		if (i >= NAND_UA_TRY_BLKS) {
			return(1);
		}

		/* OK, we have an erased block, go write pages */
		if (nand_write_page(block, 0, hdr_buf) ||
			nand_read_page(block, 0, page_buf)) {
			/* Page write failed, try next block */
			block++;
			continue;
		}

		ptr = app_code;
		for (page = hdr->blk_start_page; page <= hdr->blk_last_page; page++) {
			if (nand_write_page(block, page, ptr) ||
				nand_read_page(block, page, page_buf)) {
				/* Page write failed, try next block */
				ptr = app_code;
				block++;
				break;
			}
			ptr += nand_info.page_size;
		}

		if (ptr == app_code)
			continue;

		if ((page_count -= hdr->blk_last_page - hdr->blk_start_page + 1) <= 0)
			break;

		hdr->blk_no++;
		hdr->blk_load_addr += ptr - app_code;
		hdr->blk_start_page = 1;
		hdr->blk_last_page = page_count < (nand_info.page_cnt - 1) ?
			page_count : nand_info.page_cnt - 1;
		app_code = ptr;
		block++;
		i = 0;
	}
		
	nand_wp_set(1);
	return(0);
}

/* KSI: Beautification -- indicate blocks/pages written !!! */
/* Write bootloader to NAND */
int nand_ldr_write(void)
{
	extern u_int32_t	_text_end, _idata_start, _idata_end;
	extern u_int32_t	_ldr_entry;
	int			ldr_size;
	u_int8_t		*ldr_code, *ptr;
	u_int8_t		*hdr_buf, *page_buf;
	nand_ldr_header		*hdr;
	int			i, rc, page_count;
	
	hdr_buf = (u_int8_t *)(NAND_BUF_START);
	hdr = (nand_ldr_header *)&hdr_buf[0];
	page_buf = hdr_buf - NAND_BUF_SIZE;

	/* First, make bootloader image */
	ldr_size = _text_end - 0x20 + _idata_end - _idata_start;

	/* 
	 * Now a little bit of arithmetic. No, we can't just
	 * do something like "size / nand_info.page_size".
	 */
	switch (nand_info.page_size) {
		case SMALL_PAGE:
			page_count = ldr_size / SMALL_PAGE;
			if (ldr_size & 0x1ff)
				page_count++;
			break;
		case LARGE_PAGE:
			page_count = ldr_size / LARGE_PAGE;
			if (ldr_size & 0x7ff)
				page_count++;
			break;
		default:
			return(1);	/* Won't work with tiny page NAND */
	}

	ldr_code = ptr = page_buf - nand_info.page_size * page_count;

	for (i = 0; i < (nand_info.page_size * page_count / 4); i++)
		((u_int32_t *)ldr_code)[i] = 0xffffffff;
		
	for (i = 0x20; i < _text_end; i++)
		*ptr++ = *((u_int8_t *)i);
		
	for (i = _idata_start; i < _idata_end; i++)
		*ptr++ = *((u_int8_t *)i);

	/* Now build the header */
	for (i = 0; i < NAND_BUF_SIZE; i++)
		hdr_buf[i] = 0xff;

	hdr->magic = NAND_LDR_MAGIC;
	hdr->entry = _ldr_entry;
	hdr->num_pages = page_count;
	hdr->block = NAND_LDR_START_BLK;
	hdr->page = 1;
	
	/* OK, initial header is ready, let's start the show... */
	nand_wp_set(0);
	rc = 1;
	
	/* Main Loop */
	for (; hdr->block < (NAND_LDR_START_BLK + NAND_LDR_TRY_BLKS); hdr->block++) {
		if (nand_erase_block(hdr->block))
			continue;

		/* OK, we have an erased block, go write pages */
		if (nand_write_page(hdr->block, 0, hdr_buf) ||
			nand_read_page(hdr->block, 0, page_buf)) {
			/* Page write failed, try next block */
			continue;
		}

		ptr = ldr_code;
		for (i = 1; i <= page_count; i++) {
			if (nand_write_page(hdr->block, i, ptr) ||
				nand_read_page(hdr->block, i, page_buf)) {
				/* Page write failed, try next block */
				ptr = ldr_code;
				break;
			}
			ptr += nand_info.page_size;
		}

		if (ptr != ldr_code)
			rc = 0;
	}
		
	nand_wp_set(1);
	return(rc);
}

